#include <SDL3/SDL_events.h>
#include <SDL3/SDL_init.h>
#include <SDL3/SDL_pixels.h>
#include <SDL3/SDL_render.h>
#include <SDL3/SDL_stdinc.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#define SDL_MAIN_USE_CALLBACKS
#include <SDL3/SDL_main.h>

#include <SDL3/SDL.h>

#define WINDOW_SIZE 600
#define SWIRLINESS 35
#define ARM_COUNT 3
#define STAR_COUNT 2000

#include "../submodules/utilitiec/src/Vector/Vec2DF64.h"
#include "../submodules/utilitiec/src/rand/xoshiro256.h"

static SDL_Window *window = NULL;
static SDL_Renderer *renderer = NULL;

static SDL_Color star_colors[] = {
	{ 188,  66,  73, 255},
	{ 252, 182,  79, 255},
	{ 248, 215,  84, 255},
	{ 213, 221, 234, 255},
	{  89, 159, 195, 255},
};

static uint64_t star_color_weights[] = {
	3,
	5,
	8,
	13,
	21
};

typedef struct Star {
	Vec2DF64 position;
	SDL_Color color;
} Star;

static Star stars[STAR_COUNT];

double SDL_norm2(double x, double y)
{
	return SDL_sqrt(SDL_pow(x, 2) + SDL_pow(y, 2));
}

double Galaxy_ArmDistance(double x, double y, double swirliness, size_t arm_count)
{
	double angle = SDL_atan2(0.5 - y, 0.5 - x);
	double center_distance = SDL_norm2(x - 0.5, y - 0.5);
	double arm_distance = SDL_sin(swirliness * center_distance + angle * arm_count) + 1;

	return arm_distance / 2;
}

SDL_Color SDL_ColorMix(SDL_Color self, SDL_Color other)
{
	return (SDL_Color) {
		.r = (self.r + other.r) / 2,
		.g = (self.g + other.g) / 2,
		.b = (self.b + other.b) / 2,
		.a = (self.a + other.a) / 2,
	};
}

SDL_Color SDL_ColorLerp(SDL_Color self, SDL_Color other, uint progress)
{
	uint inverse_progress = 100 - progress;
	return (SDL_Color) {
		.r = (self.r * inverse_progress + other.r * progress) / 100,
		.g = (self.g * inverse_progress + other.g * progress) / 100,
		.b = (self.b * inverse_progress + other.b * progress) / 100,
		.a = (self.a * inverse_progress + other.a * progress) / 100,
	};
}


size_t _SelectWeighed(uint64_t weights[], size_t weight_count, Xoshiro256State* state)
{
	uint64_t sum = 0;
	for (size_t i = 0; i < weight_count; i++) sum += weights[i];

	uint64_t chosen = xoshiro256_next(state) % sum;

	for (size_t i = 0; i < weight_count; i++) {
		if (chosen < weights[i]) return i;
		else chosen -= weights[i];
	}

	return SIZE_MAX;
}

void GenerateStars()
{
	Xoshiro256State random_state = { 1, 2, 3, 4 };
	for (size_t i = 0; i < SDL_arraysize(stars); i++) {
		Star* star = stars + i;

		double arm_distance;
		uint64_t discard_probability;
		do {
			star->position.x = xoshiro256_next(&random_state) % WINDOW_SIZE;
			star->position.y = xoshiro256_next(&random_state) % WINDOW_SIZE;
			size_t color_index = _SelectWeighed(
				star_color_weights, 
				SDL_arraysize(star_color_weights), 
				&random_state
			);
			star->color = star_colors[color_index];

			arm_distance = Galaxy_ArmDistance(
				star->position.x / WINDOW_SIZE,
				star->position.y / WINDOW_SIZE,
				SWIRLINESS,
				ARM_COUNT
			);
			discard_probability = xoshiro256_next(&random_state) % 1000;
		} while (arm_distance * 1000 > discard_probability);
	}
}

void SDL_RenderStar(Star* star)
{
	SDL_SetRenderDrawColor(renderer, star->color.r, star->color.g, star->color.b, star->color.a);
	SDL_RenderPoint(renderer, star->position.y, star->position.x);

	SDL_Color white = { 255 };
	SDL_Color adjacent_color = SDL_ColorLerp(star->color, white, 5); 
	SDL_SetRenderDrawColor(
		renderer,
		adjacent_color.r,
		adjacent_color.g,
		adjacent_color.b,
		adjacent_color.a
	);
	SDL_RenderPoint(renderer, star->position.y - 1, star->position.x);
	SDL_RenderPoint(renderer, star->position.y + 1, star->position.x);
	SDL_RenderPoint(renderer, star->position.y, star->position.x - 1);
	SDL_RenderPoint(renderer, star->position.y, star->position.x + 1);
}

void SDL_RenderArms()
{
	for (size_t y = 0; y < WINDOW_SIZE; y++) {
		double norm_y = (double) y / WINDOW_SIZE;
		for (size_t x = 0; x < WINDOW_SIZE; x++) {
			double norm_x = (double) x / WINDOW_SIZE;
			double distance = Galaxy_ArmDistance(norm_x, norm_y, SWIRLINESS, ARM_COUNT);
			uint progress = distance * 100;
			SDL_Color black = { 0, 0, 0, 255 };
			SDL_Color grey = { 31, 31, 31, 255 };
			SDL_Color arm_color = SDL_ColorLerp(black, grey, 101 - progress);
			SDL_SetRenderDrawColor(renderer, arm_color.r, arm_color.g, arm_color.b, arm_color.a);
			SDL_RenderPoint(renderer, y, x);
		}
	}
}

SDL_AppResult SDL_AppInit(void **appstate, int argc, char **argv)
{
	GenerateStars();

	if (!SDL_CreateWindowAndRenderer("Spiral Galaxy", WINDOW_SIZE, WINDOW_SIZE, 0, &window, &renderer)) {
		SDL_Log("Couldn't create window and renderer: %s\n", SDL_GetError());
		return SDL_APP_FAILURE;
	}

	SDL_SetRenderDrawColor(renderer, 0, 0, 0, 0);
	SDL_RenderClear(renderer);

	SDL_RenderArms();
	for (size_t i = 0; i < SDL_arraysize(stars); i++) {
		SDL_RenderStar(stars + i);
	}

	SDL_RenderPresent(renderer);


	return SDL_APP_CONTINUE;
}

SDL_AppResult SDL_AppEvent(void *appstate, SDL_Event *event)
{
	switch (event->type) {
		case SDL_EVENT_QUIT:
			return SDL_APP_SUCCESS;

		default:
			break;
	}

	return SDL_APP_CONTINUE;
}

SDL_AppResult SDL_AppIterate(void *appstate)
{
	return SDL_APP_CONTINUE;
}

void SDL_AppQuit(void *appstate, SDL_AppResult result)
{
}
